(*  :Title:	Cost Measures for Analog Filter Time and Frequency Response *)

(*  :Authors:	Brian L. Evans and Niranjan Damera-Venkata *)

(*
    :Summary:   This file defines functions to measure an analog filter
		design in terms of its magnitude error, deviation from
		linear phase, filter quality, and percent overshoot.   The
		complete cost function is a weighted measure of the
		four types of cost.
 *)

(*  :Context:	FilterOptimization`CostMeasures`  *)

(*  :PackageVersion:  FOP 1.1	@(#)CostMeasures.m	1.10	06/21/98  *)

(*
   :Copyright:

Copyright (c) 1993-1998 The Regents of the University of California.
All rights reserved.

Permission is hereby granted, without written agreement and without
license or royalty fees, to use, copy, modify, and distribute this
software and its documentation for any purpose, provided that the above
copyright notice and the following two paragraphs appear in all copies
of this software.

IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY 
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES 
ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF 
THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF 
SUCH DAMAGE.

THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS.
 *)

(*  :History:	*)

(*  :Keywords:	*)

(*
    :Source:
 *)

(*  :Warning:	*)

(*  :Mathematica Version:  2.0  *)

(*  :Limitation:  *)

(*
    :Discussion:
   
	cost functions for an analog filter time and frequency response

	free parameters --  passed as arguments to individual functions:

	wp	  passband frequency
	ws	  stopband frequency
	wmax	  maximum linear phase frequency 
	tpeak	  time at which peak occurs in step response
*)

(*  :Functions:	*)


BeginPackage[ "FilterOptimization`CostMeasures`",
	      "FilterOptimization`Filters`",
	      "FilterOptimization`CodeGen`" ]


(*  U S A G E     I N F O R M A T I O N  *)

w::usage =
"w is the reserved frequency variable in units of rad/sec.";

$StopbandDecades::usage =
"$StopbandDecades is the number decades of frequency beyond the
the stopband frequency over which to compute the magnitude error
in the stopband.";

linearPhaseCost::usage =
"linearPhaseCost[ r, i, c, d, n, m, wmax ] computes the cost of the deviation
from linear phase over the range of frequencies from 0 to wmax.
Usually, wmax is set to be the frequency at the end of the
passband, wp. \
The measure is normalized by dividing by wmax.";

magnitudeCostPassband::usage =
"magnitudeCostPassband[ r, i, c, d, n, m, wp, ws ] computes the least-squares
error in the passband of the magnitude response of an analog filter
with n pole pairs and m zero pairs.
The passband is defined as those frequencies up to wp,
the passband frequency.
The measure is normalized by dividing by wp.
The function also takes two optional arguments---
dp the deviation in magnitude response in the passband and
ds the deviation in magnitude response in the stopband.";

magnitudeCostStopband::usage =
"magnitudeCostStopband[ r, i, c, d, n, m, wp, ws ] computes the least-squares
error in the stopband of the magnitude response of an analog filter
with n pole pairs.
The stopband is defined as those frequencies beyond ws,
the stopband frequency.
The actual interval taken is ws to 10^($StopbandDecades) ws.
The measure is normalized by dividing by
ws (10^($StopbandDecades) - 1).
The function also takes two optional arguments---
dp the deviation in magnitude response in the passband and
ds the deviation in magnitude response in the stopband.";

magnitudeCostTransband::usage =
"magnitudeCostTransband[ r, i, c, d, n, m, wp, ws ] computes the least-squares
error in the transition band of the magnitude response of an analog
filter with n pole pairs.
The transition band is defined as those frequencies between
the passband frequency ws and the stopband frequency wp
The measure is normalized by (ws - wp).
The function also takes two optional arguments---
dp the deviation in magnitude response in the passband and
ds the deviation in magnitude response in the stopband.";

overshootCost::usage =
"overshootCost[ r, i, c, d, n, m, tpeak ] penalizes the percent overshoot
in the analog filter with 2 n poles at r[j] +/- i[j] for j = 1 .. n.
The parameter tpeak specifies at which time the overshoot occurs.";

qualityCost::usage =
"qualityCost[ r, i, n ] penalizes analog filters with high quality
factors.
Analog filters with high quality factors are prone to oscillation
and are not robust against noisy inputs.
The analog filter has n pole pairs.";

weightedCosts::usage =
"weightedCosts[ weights, r, i, n, wp, ws, wmax, tpeak ]
computes the weighted cost of the design characteristics of an
analog filter with n pole pairs.
The function also takes two optional arguments---
dp the deviation in magnitude response in the passband and
ds the deviation in magnitude response in the stopband.
The filter has passband frequency wp and stopband frequency ws.
It is desired that the filter exhibit linear phase from DC to wmax.
The parameter tpeak specifies when the peak overshoot occurs.
Of the six weights, the first three weight the deviation from ideal
of the magnitude response in the passband, transition region, and
stopband, respectively.
The remaining three weight the phase, quality, and overshoot costs.";

(*  E N D     U S A G E     I N F O R M A T I O N  *)


Begin["`Private`"]


(*  Settings  *)

$StopbandDecades = 2	(* # of stopband decades over which to compute error *)


(*  Rules to simplify forms involving arctangent functions not in 2.0 or 2.1 *)

arctanRules = {
        ArcTan[x_, 0] :> 0,
        Cos[ ArcTan[x_, y_] ] :> x / Sqrt[x^2 + y^2],
        Sin[ ArcTan[x_, y_] ] :> y / Sqrt[x^2 + y^2]
}

If [ TrueQ[ $VersionNumber <= 2.2 ],
     arctanRules = arctanRules ~Join~ {
	Cos[ ArcTan[z_] ] :> 1 / Sqrt[1 + z^2],
        Sin[ ArcTan[z_] ] :> z / Sqrt[1 + z^2] } ]

If [ TrueQ[ $VersioNumber >= 2.0 ],
     otherRules = { (x_^r_)^q_ :> x /; r q == 1 },
     otherRules = { } ]

simpleRules = Join[arctanRules, otherRules]


(*  weighted cost function  *)

(* Extended for zeros *)

weightedCosts[ { wMagPassband_, wMagTransband_, wMagStopband_, 
		 wPhase_, wQuality_, wOvershoot_ },
	r_, i_, c_, d_, n_, m_, wp_, ws_, wmax_, tpeak_, dp_:0, ds_:0 ] :=
( wMagPassband * magnitudeCostPassband[ r, i, c, d, n, m, wp, ws, dp, ds ] +
wMagTransband * magnitudeCostTransband[ r, i, c, d, n, m, wp, ws, dp, ds ] +
wMagStopband * magnitudeCostStopband[ r, i, c, d, n, m, wp, ws, dp, ds ] +
wPhase * linearPhaseCost[ r, i, c, d, n, m, wmax ] +
wQuality * qualityCost[ r, i, n ] +
wOvershoot * overshootCost[ r, i, c, d, n, m, tpeak ] ) /. simpleRules

(* magnitudeCostPassband  *)
magnitudeCostPassband[ r_, i_, c_, d_, n_, m_, wp_, ws_, dp_:0, ds_:0 ] :=
	Block [ {norm},
		norm = 1/wp;
                Print["magnitude cost passband"];
	norm integrate[ ( magnitude[r, i, c, d, n, m,  w] - (1 - dp/2) )^2,
				{w, 0, wp} ] ]

(* magnitudeCostTransband *)
(* -- mag. response in the transition band is modelled as a line m x + b *)
magnitudeCostTransband[ r_, i_, c_, d_, n_, m_, wp_, ws_, dp_:0, ds_:0 ] :=
	Block [ {b, sl, norm},
		sl = (1 - dp/2) / (wp - ws);
		slb = -sl ws;
		norm = 1 / (ws - wp);
               Print["magnitudeTransband"];
	norm integrate[ ( magnitude[r, i, c, d, n, m, w] - sl w - slb )^2,
				{w, wp, ws} ] ]

(* magnitudeCostStopband *)
magnitudeCostStopband[ r_, i_, c_, d_, n_, m_, wp_, ws_, dp_:0, ds_:0 ] :=
	Block [ {norm, wsmax},
		wsmax = 10^( $StopbandDecades ) ws;
		norm = 1 / ( ws ( 10^($StopbandDecades) - 1) );
                Print["magnitudeStopband"];
 norm integrate[ magnitude[r, i, c, d, n, m, w]^2, {w, ws, wsmax} ] ]

(* linearPhaseCost *)
linearPhaseCost[ r_, i_, c_, d_, n_, m_, wmax_ ] :=
	Block [	{norm, slope},
		slope = linearPhaseSlope[r, i, c, d, n, m, wmax];
		norm = 1 / wmax;

                Print["Linear phase cost"];
		norm integrate[ ( phase[r, i, c, d, n, m, w] - slope w )^2,
			{w, 0, wmax} ] ]

(* qualityCost -- minimum value for quality is 0.5  *)
qualityCost[ r_, i_, n_ ] := 
	Block [ {}, quality[r, i, n] - 0.5 ]

(* overshootCost *)
overshootCost[ r_, i_, c_, d_, n_, m_, tpeak_ ] :=
	Block [	{}, Print["overshoot cost"];
       100 ( stepResponse[ r, i, c, d, n, m, tpeak ] - 1 )^2 ]


End[]
EndPackage[]

Protect[ linearPhaseCost,	magnitudeCostPassband,
	 magnitudeCostStopband,	magnitudeCostTransband,
	 overshootCost,		qualityCost,
	 w,			weightedCosts ];

Protect[ $StopbandDecades ];

Null
